# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'timestamp_dialog_alt.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_manualTimestampDialog(object):
    def setupUi(self, manualTimestampDialog):
        manualTimestampDialog.setObjectName("manualTimestampDialog")
        manualTimestampDialog.resize(863, 544)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(manualTimestampDialog.sizePolicy().hasHeightForWidth())
        manualTimestampDialog.setSizePolicy(sizePolicy)
        manualTimestampDialog.setSizeGripEnabled(True)
        self.gridLayout = QtWidgets.QGridLayout(manualTimestampDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.label_13 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_13.setObjectName("label_13")
        self.horizontalLayout_6.addWidget(self.label_13)
        self.gridLayout.addLayout(self.horizontalLayout_6, 0, 0, 1, 1)
        self.verticalLayout_10 = QtWidgets.QVBoxLayout()
        self.verticalLayout_10.setObjectName("verticalLayout_10")
        self.label_14 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_14.setObjectName("label_14")
        self.verticalLayout_10.addWidget(self.label_14)
        self.radioButtonNTSC = QtWidgets.QRadioButton(manualTimestampDialog)
        self.radioButtonNTSC.setChecked(True)
        self.radioButtonNTSC.setObjectName("radioButtonNTSC")
        self.verticalLayout_10.addWidget(self.radioButtonNTSC)
        self.radioButtonPAL = QtWidgets.QRadioButton(manualTimestampDialog)
        self.radioButtonPAL.setObjectName("radioButtonPAL")
        self.verticalLayout_10.addWidget(self.radioButtonPAL)
        self.line_2 = QtWidgets.QFrame(manualTimestampDialog)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.verticalLayout_10.addWidget(self.line_2)
        self.radioButtonCustom = QtWidgets.QRadioButton(manualTimestampDialog)
        self.radioButtonCustom.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.radioButtonCustom.setObjectName("radioButtonCustom")
        self.verticalLayout_10.addWidget(self.radioButtonCustom)
        self.gridLayout.addLayout(self.verticalLayout_10, 3, 0, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_9 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_9.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout_3.addWidget(self.label_9)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(manualTimestampDialog)
        self.label.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.frameNum1 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.frameNum1.setObjectName("frameNum1")
        self.verticalLayout.addWidget(self.frameNum1)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label_2 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_2.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_2.addWidget(self.label_2)
        self.hh1 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.hh1.setObjectName("hh1")
        self.verticalLayout_2.addWidget(self.hh1)
        self.horizontalLayout.addLayout(self.verticalLayout_2)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_3 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_3.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_3.addWidget(self.label_3)
        self.mm1 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.mm1.setObjectName("mm1")
        self.verticalLayout_3.addWidget(self.mm1)
        self.horizontalLayout.addLayout(self.verticalLayout_3)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setSpacing(0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label_4 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_4.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_4.addWidget(self.label_4)
        self.ss1 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.ss1.setObjectName("ss1")
        self.verticalLayout_4.addWidget(self.ss1)
        self.horizontalLayout.addLayout(self.verticalLayout_4)
        self.horizontalLayout_3.addLayout(self.horizontalLayout)
        self.gridLayout.addLayout(self.horizontalLayout_3, 1, 0, 1, 1)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_10 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_10.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft)
        self.label_10.setObjectName("label_10")
        self.horizontalLayout_4.addWidget(self.label_10)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout()
        self.verticalLayout_5.setSpacing(0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label_5 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_5.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft)
        self.label_5.setObjectName("label_5")
        self.verticalLayout_5.addWidget(self.label_5)
        self.frameNum2 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.frameNum2.setObjectName("frameNum2")
        self.verticalLayout_5.addWidget(self.frameNum2)
        self.horizontalLayout_2.addLayout(self.verticalLayout_5)
        self.verticalLayout_6 = QtWidgets.QVBoxLayout()
        self.verticalLayout_6.setSpacing(0)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.label_6 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_6.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_6.setObjectName("label_6")
        self.verticalLayout_6.addWidget(self.label_6)
        self.hh2 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.hh2.setObjectName("hh2")
        self.verticalLayout_6.addWidget(self.hh2)
        self.horizontalLayout_2.addLayout(self.verticalLayout_6)
        self.verticalLayout_7 = QtWidgets.QVBoxLayout()
        self.verticalLayout_7.setSpacing(0)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.label_7 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_7.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_7.setObjectName("label_7")
        self.verticalLayout_7.addWidget(self.label_7)
        self.mm2 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.mm2.setObjectName("mm2")
        self.verticalLayout_7.addWidget(self.mm2)
        self.horizontalLayout_2.addLayout(self.verticalLayout_7)
        self.verticalLayout_8 = QtWidgets.QVBoxLayout()
        self.verticalLayout_8.setSpacing(0)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.label_8 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_8.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignHCenter)
        self.label_8.setObjectName("label_8")
        self.verticalLayout_8.addWidget(self.label_8)
        self.ss2 = QtWidgets.QLineEdit(manualTimestampDialog)
        self.ss2.setObjectName("ss2")
        self.verticalLayout_8.addWidget(self.ss2)
        self.horizontalLayout_2.addLayout(self.verticalLayout_8)
        self.horizontalLayout_4.addLayout(self.horizontalLayout_2)
        self.gridLayout.addLayout(self.horizontalLayout_4, 2, 0, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(manualTimestampDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 5, 0, 1, 1)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_11 = QtWidgets.QLabel(manualTimestampDialog)
        self.label_11.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.label_11.setObjectName("label_11")
        self.horizontalLayout_5.addWidget(self.label_11)
        self.verticalLayout_9 = QtWidgets.QVBoxLayout()
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.frameDeltaTime = QtWidgets.QLineEdit(manualTimestampDialog)
        self.frameDeltaTime.setCursorMoveStyle(QtCore.Qt.VisualMoveStyle)
        self.frameDeltaTime.setObjectName("frameDeltaTime")
        self.verticalLayout_9.addWidget(self.frameDeltaTime)
        self.horizontalLayout_5.addLayout(self.verticalLayout_9)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem)
        self.gridLayout.addLayout(self.horizontalLayout_5, 4, 0, 1, 1)

        self.retranslateUi(manualTimestampDialog)
        self.buttonBox.accepted.connect(manualTimestampDialog.accept)
        self.buttonBox.rejected.connect(manualTimestampDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(manualTimestampDialog)
        manualTimestampDialog.setTabOrder(self.frameNum1, self.hh1)
        manualTimestampDialog.setTabOrder(self.hh1, self.mm1)
        manualTimestampDialog.setTabOrder(self.mm1, self.ss1)
        manualTimestampDialog.setTabOrder(self.ss1, self.frameNum2)
        manualTimestampDialog.setTabOrder(self.frameNum2, self.hh2)
        manualTimestampDialog.setTabOrder(self.hh2, self.mm2)
        manualTimestampDialog.setTabOrder(self.mm2, self.ss2)
        manualTimestampDialog.setTabOrder(self.ss2, self.frameDeltaTime)

    def retranslateUi(self, manualTimestampDialog):
        _translate = QtCore.QCoreApplication.translate
        manualTimestampDialog.setWindowTitle(_translate("manualTimestampDialog", "Manual Timestamp Entry"))
        self.label_13.setText(_translate("manualTimestampDialog", "<html><head/><body><p>PyOTE assumes that a frame timestamp represents the UTC time of the synch pulse that <br/>first appears in the composite video output as the frame/field is emitted from the camera. <br/><br/>For an IOTA-VTI frame display, this will be the earliest of the two times displayed in the overlay text.</p><p>Note: if the light curve was processed in field mode, frame numbers are still used but become fractional<br/>to indicate fields.  And the frame time remains the proper time unit to specify.<br/></p></body></html>"))
        self.label_14.setText(_translate("manualTimestampDialog", "<html><head/><body><p><br/>In order to detect dropped readings, PyOTE needs to know the expected frame time. Make a selection below:<br/></p></body></html>"))
        self.radioButtonNTSC.setText(_translate("manualTimestampDialog", "use NTSC frame time (1.001/30 seconds)"))
        self.radioButtonPAL.setText(_translate("manualTimestampDialog", "use PAL    frame time (1.000/25 seconds)"))
        self.radioButtonCustom.setText(_translate("manualTimestampDialog", "use Custom frame time (enter number or numerical expression in box below)"))
        self.label_9.setText(_translate("manualTimestampDialog", "<html><head/><body><p>Timestamp (early frame)</p></body></html>"))
        self.label.setText(_translate("manualTimestampDialog", "frame number"))
        self.label_2.setText(_translate("manualTimestampDialog", "hh"))
        self.label_3.setText(_translate("manualTimestampDialog", "mm"))
        self.label_4.setText(_translate("manualTimestampDialog", "ss.ssss"))
        self.label_10.setText(_translate("manualTimestampDialog", "<html><head/><body><p>Timestamp (later frame)</p></body></html>"))
        self.label_5.setText(_translate("manualTimestampDialog", "frame number"))
        self.label_6.setText(_translate("manualTimestampDialog", "hh"))
        self.label_7.setText(_translate("manualTimestampDialog", "mm"))
        self.label_8.setText(_translate("manualTimestampDialog", "ss.ssss"))
        self.label_11.setText(_translate("manualTimestampDialog", "<html><head/><body><p>Custom frame time (seconds)</p></body></html>"))

